/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#pragma once

#include "Linker.h"
#include "instructions/Instruction.h"

class Assembler {
private:
    const Linker &linker;
    bool padEndFunction = true;
    std::vector<uint8_t> bytes;
    std::unordered_map<std::string, uint32_t> knownRegisters;
    std::unordered_map<std::string, std::pair<uint32_t, bool>> entries;

    uint32_t placeFunction(uint32_t base, const Function &function);

    const Function *getFunctionByLabel(const std::string &l);

public:

    Assembler(const Linker &linker) : linker(linker) {

    }

    //Default max size 1 MiB
    void assemble(uint32_t baseAddress = 0x080000, uint32_t imageSize = 0x100000);

    void setSignature(const std::string &str);

    void setDoFunctionPadding(bool pad) {
        padEndFunction = pad;
    }

    const auto &getImage() const {
        return bytes;
    }

    enum class FunctionPlacementMode {
        PlaceAll,
        PlaceOnUse,
    };


    void setFunctionPlacementMode(FunctionPlacementMode mode) {
        placementMode = mode;
    }

private:
    FunctionPlacementMode placementMode = FunctionPlacementMode::PlaceOnUse;
};
