/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#pragma once

#include "../Instruction.h"
#include "Stack.h"

namespace Instruction {


    class Call : public Instruction {
    private:
        AddressWithOffset addressWithOffset;
        uint32_t address;
    public:

        explicit Call(const AddressWithOffset &address);
        explicit Call(uint32_t directAddress);

        CommandOp getCommand() const override {
            return CommandOp::END;
        };

        std::vector<uint8_t> toBytes() const override;

        size_t getByteSize() const override{
            //TODO: do not hardcode this
            return 4 + 8 + 12; //4 bytes for Return/Jump + 8 bytes for Write + 12 for CalcImm
        }

        static std::unique_ptr<Instruction> fromTokens(const Instruction::parseOperatorResult &op, const std::vector<Token> &tokens);
    };
    class Ret : public Instruction {
    private:
        std::vector<std::unique_ptr<Instruction>> instructions;
    public:

        explicit Ret();

        CommandOp getCommand() const override {
            return CommandOp::END;
        };

        std::vector<uint8_t> toBytes() const override;

        size_t getByteSize() const override{
            size_t s = 0;
            for(auto& i : instructions){
                s += i->getByteSize();
            }
            return s;
        }

        static std::unique_ptr<Instruction> fromTokens(const Instruction::parseOperatorResult &op, const std::vector<Token> &tokens);
    };


}